/*
 * decl.h -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) 2005, Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

extern uint8_t		verbose;
extern uint32_t		debug;
extern uint32_t		last_progID;
extern struct equiptable_2000access	eqAC;
extern struct equiptable_2000F		eqF;
extern struct equiptable_2000C		eqC;
extern struct equiptable_2000E		eqE;
extern struct DIRECtable_2000access    Dt2000access;
extern struct NONSHAREDEVICEtable_2000access NSDEV2000access;
extern struct Program_Entry		Pentry;

extern uint8_t		IDtable[1024 * 80];
extern uint32_t		IDtablesize;
extern uint32_t		tape_mach_type;
extern uint8_t		extract;
extern uint8_t		create;
extern uint8_t		tokenize;
extern uint8_t		writedir;
extern uint8_t		overwrite;
extern uint8_t		convert;
extern uint32_t		quoteflag;
extern uint8_t		display;
extern uint8_t		only_programs;
extern uint8_t		only_files;
extern uint8_t		only_ascii_files;
extern uint8_t *	feature_level_name[];
extern uint8_t *	file_flags[];
extern uint8_t *	system_level_name[];
extern uint32_t		nentries;
extern uint8_t		rentry[][512];

extern char		*progname;
extern char		Nfilename[];
extern char		infilename[10][MAXPATHLEN];
extern char		buf[];

/* Global function declarations */
void load_program_entry (uint32_t, uint8_t *, uint32_t, uint32_t);
void convert_program (struct Program_Entry  *);
void convert_file (struct Program_Entry  *);
void convert_asciifile (struct Program_Entry  *);
/* PRINTFLIKE */
void dprintf(uint32_t, char *, ...);
/* PRINTFLIKE */
void lprintf (char *, char *, ...);
void hexdump (uint8_t *, uint32_t);
uint8_t * machtypename (uint32_t);
/* PRINTFLIKE */
void eprintf (char *, ...);
void day2date (uint32_t, uint32_t, uint32_t);
void eval_2000access_equipment_table (uint8_t *, uint32_t);
void eval_2000F_equipment_table (uint8_t *, uint32_t);
void eval_2000C_equipment_table (uint8_t *, uint32_t);
void eval_2000E_equipment_table (uint8_t *, uint32_t);
void eval_ACCESS_DIREC_table (uint8_t *, uint32_t);
void eval_2000C_DIREC_table (uint8_t *, uint32_t);
void eval_2000access_NONSHAREDEVICE_table (uint8_t *, uint32_t);
void load_ID_table (uint8_t *, uint32_t, uint32_t, uint32_t *);
void dump_ID_table (uint8_t *, uint32_t, uint32_t);
uint32_t HPputchar (uint8_t *, uint8_t *, uint32_t, uint32_t *);
char * ftoHP (float, char *);
float HPftoIEEE (uint32_t);
uint32_t IEEEtoHPf (float);
uint8_t * printarrow (uint8_t *, uint8_t *, uint32_t, uint32_t);
void add_line (struct line_table **, uint32_t, uint8_t *);
/* PRINTFLIKE */
void errorexit (int, char *, ...);
void extract_entry (struct Program_Entry *, char *);
uint32_t wildcmp (uint8_t *, uint8_t *);
uint32_t check_entrylist (uint8_t *, uint8_t *list, uint32_t, uint32_t);
void * n_malloc(int, char *, int);
void n_free(void *, char *, int);
void n_list(void);
uint32_t	validate_id (uint32_t);
uint32_t	validate_filnam (uint8_t *, uint32_t);
void		tsb_create(void);
uint8_t *	get_number(uint8_t *, uint32_t, void *);
uint8_t *	get_next_character (uint8_t *);
uint8_t	*	get_subscript (uint8_t **, uint32_t *, uint8_t *, uint32_t *);
uint8_t	*	get_nth_character (uint8_t *, uint32_t);
uint8_t	*	get_source_string (uint8_t *, uint8_t *, uint8_t *, uint8_t *, uint32_t *);
void		tsb_tokenize(void);
uint32_t	tokenize_line (uint8_t *, struct Program_Space *);

