/*
 * tokenize.c -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) 2005, Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

/*
 * TODO:
 * o	Error message format:
 *		ERROR: 320 READ #(N1+1)5
 *		ERROR:                 ^
 *		ERROR: Missing semi-colon after file reference.
 */

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/types.h>

#include "config.h"

/*#include "tsb.h"*/
#include "hptsb.h"
#include "hptoken.h"
#include "decl.h"

uint32_t demand_string (uint8_t  *, uint32_t, uint32_t *, uint8_t *);

uint8_t *ans_formula (uint8_t **, uint32_t *, uint8_t *, uint32_t *, uint8_t *, uint8_t *, uint8_t *);
uint8_t *get_simple_variable (uint8_t *lp, uint8_t *, uint8_t *);
uint8_t *get_string_variable (uint8_t *, uint8_t *, uint8_t *);
uint8_t *get_string_literal (uint8_t *, uint8_t *, uint32_t *);
uint8_t *get_relational_operator (uint8_t *, uint32_t *);

#define TOKERROR(x,z) \
	{	ps->error_pos = lp; \
		ps->error_text = z; \
		return (x); \
	}

#ifdef OLD
/* HACKS FOR LINKING */

#define MAXENTRIES      512
 
/*uint32_t        last_progID = 0;*/

/*uint32_t        nentries = 0;*/
uint8_t         rentry[MAXENTRIES][512];

/* END HACKS */
#endif /* OLD */

uint8_t * get_nth_character (uint8_t *, uint32_t);
uint8_t * eval_integer (uint8_t *, uint32_t *);

uint32_t
tokenize_line (uint8_t * line, struct Program_Space *ps)
{
int32_t		msflag, maxcr, pflag, sflag, dflag;
int32_t		stlen;
	uint8_t		*lp = line;
	uint8_t		*nlp;
	uint8_t		*mp;
	uint8_t		*m;
	uint8_t		*sp;
	uint8_t		*st;
	uint8_t		*stmt;
	uint8_t		*nstmt;
	uint8_t		varname;
	uint8_t		varnum;
	uint8_t		vartype;
	uint32_t	stringlen;
	uint8_t		fixstringlen;
	uint8_t		litstring[512];
	uint32_t	dimvar, dim1 = 0, dim2 = 0;

	uint32_t	lineno = 0;
	uint32_t	wlineno;
	uint16_t	ilineno;
	uint32_t	sttype;
	uint32_t	save_sttype;
	uint32_t	new_sttype;
	uint32_t	fileref;
	uint8_t		*statement;

	float		wnumber;
	uint32_t	HPfloat;

	if (ps != NULL) {
		m = ps->mem;
		mp = m + ps->len;
		st = ps->symtab;
		sp = st + ps ->symtabsize;
	}

	statement = ps->statement;

	/* Determine sequence number */

eprintf ("Tokenizing line: %s\n", lp);
	if ((nlp = eval_integer (lp, &lineno)) == NULL)
		TOKERROR (1, "Invalid statement number")
	lp = nlp;

	lp = match_token (lp, &sttype, Bst_ACCESS, Bst_ACCESS_count, 0x8000);
	if (sttype != 044 && sttype != 051) {
		/* Only skip blanks after the statement type if not REM or IMAGE */
		lp = get_next_character (lp);
	}

	if (sttype == 9999)	/* Normally, implied let matches automatically */
		sttype = 073;
	if (sttype < 0)
		TOKERROR (2, "UNKNOWN ERROR")

	dprintf (DEBUG_TOKEN, "*** Statement Type is: (%o) %s\n",
			sttype, Bst_ACCESS[sttype].ascii);

	msflag = -9;
	maxcr = 3;
	pflag = 0;
	sflag = 0;
	dflag = 0;

	/*
	 * Statement Syntax Parsing.
	 */

	statement[0] = lineno >> 8;	/* Statement No, upper byte */
	statement[1] = lineno & 0xff;	/* Statement No, lower byte */
	/* statement[2] and statement[3] (stlen) are set at the very end of parsing */
	stmt = &statement[4];
	stlen = 2;

	switch (sttype) {
	case 031:	/* LIB?   Special Case for TSC */
		lp = get_next_character (lp);
		break;
	case 032:	/* SYSTEM */
		lp = get_next_character (lp);
		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable missing.")
		}
		lp = nlp;
		stlen ++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		sttype = 0;
		nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript.")
		stmt = nstmt;

		if (*lp != ',')
			TOKERROR (50, "Missing comma")
		lp = get_next_character (lp + 1);
		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
		sttype = 2;
		if (stringlen != 9999) {
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9));
			*stmt++ = ITOWL ((sttype << 9));
			sttype = 1;
			stlen ++;
			*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
			*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
			memcpy (stmt, litstring, stringlen);
			stlen += (stringlen + 1) / 2;
			stmt += stringlen + (stringlen % 2);
			dprintf (DEBUG_TOKEN, "STRINGIT2: Stored a string literal of length %d\n", stringlen);
		} else if (lp != nlp) {
			lp = nlp;
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

			dprintf(DEBUG_TOKEN, "STRINGIT2: Stored a string variable, varname=%d  vartype=%d, stringlen=%d.\n", varname, vartype, stringlen);
			sttype = 0;
			nstmt = get_subscript (&nlp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			lp = nlp;
			stmt = nstmt;

		} else
			TOKERROR (49, "Missing string expression")
		lp = nlp;
		break;

	case 033:	/* CONVERT */

		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp != nlp) {
			lp = nlp;
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			dprintf(DEBUG_TOKEN, "STRINGIT2: Stored a string variable, varname=%d  vartype=%d, stringlen=%d.\n", varname, vartype, stringlen);
			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
		} else if ((nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen)) != lp && stringlen != 9999) {
			lp = nlp;
			if (sttype) {
				stlen ++;
				*stmt++ = ITOWH ((sttype << 9));
				*stmt++ = ITOWL ((sttype << 9));
			}
			sttype = 1;
			fixstringlen = stringlen;

			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | (fixstringlen & 0377));
			*stmt++ = ITOWL ((sttype << 9) | (fixstringlen & 0377));

			memcpy (stmt, litstring, stringlen);
			stlen = stlen + (stringlen + 1) / 2;
			stmt += stringlen;
			if (stringlen % 2) {
				*stmt++ = 0;
			}

			sttype = 0;
		} else if ((nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum)) != stmt) {
			stmt = nstmt;
#ifdef OLD
			if (varname != 0) {
				stlen++;
				*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				varname = 0;
				vartype = 0;

				sttype = 0;
				nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;
			}
#endif
		} else
			TOKERROR (74, "Bad or missing CONVERT source")

		if ((nlp = match_literal (lp, "TO")) == NULL)
			TOKERROR (44, "Missing or illegal TO")
		lp = nlp;

		sttype = 076;
				
		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable missing")
		}
		lp = nlp;
		stlen ++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		sttype = 0;
		nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		if (*lp != ',')
			break;
		sttype = 2;
		lp = get_next_character (lp + 1);

		/* Demand a statement number now */
		if ((lp = get_number (lp, GN_LINENO, &wlineno)) == NULL)
			TOKERROR (23, "Illegal or missing statement number for exception after comma")
		stlen += 1;
		*stmt++ = ITOWH (0x8000 | (sttype << 9) | 3);
		*stmt++ = ITOWL (0x8000 | (sttype << 9) | 3);
		ilineno = wlineno;
		stlen += 1;
		*stmt++ = ITOWH (ilineno);
		*stmt++ = ITOWL (ilineno);

		break;

	case 034:	/* LOCK */
	case 035:	/* UNLOCK */
		stlen += 1;
		*stmt++ = ITOWH (sttype << 9);
		*stmt++ = ITOWL (sttype << 9);
		lp = get_next_character (lp);
		if (*lp != '#')
			TOKERROR (21, "Missing '#' for file reference")
		lp = get_next_character(lp + 1);

		sttype = 4;
		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (30, "Missing file reference")
		stmt = nstmt;
		

		lp = get_next_character (lp);
		if (*lp != ',')
			break;
		lp = get_next_character (lp + 1);

		lp = get_simple_variable (lp, &varname, &vartype);
		if (!varname)
			TOKERROR (31, "Missing variable")
		sttype = 2;
		stlen ++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		
		lp = get_next_character (lp);
		sttype = 0;
		nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;
		break;
	case 036:	/* CREATE */
	case 037:	/* PURGE */
		save_sttype = sttype;
		lp = get_simple_variable (lp, &varname, &vartype);
		if (!varname)
			TOKERROR (51, "Missing return variable")
		stlen ++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		sttype = 0;
		nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		if (*lp != ',')
			TOKERROR (31, "Variable expected")
		lp = get_next_character (lp + 1);

		sttype = 2;
		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
		dprintf (DEBUG_TOKEN, "get_source_string: RETURNED:  lp=%x  nlp=%x, varname=%d  vartype=%d,  litstring5='%-.5s' stringlen=%d\n",
				lp, nlp, varname, vartype, litstring, stringlen);
		lp = nlp;
		if (stringlen == 9999) {
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
			dprintf(DEBUG_TOKEN, "STRINGIT: Stored a string variable.\n");
		} else {
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9));
			*stmt++ = ITOWL ((sttype << 9));
			sttype = 1;
			stlen ++;
			*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
			*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
			memcpy (stmt, litstring, stringlen);
			stlen += (stringlen + 1) / 2;
			stmt += stringlen + (stringlen % 2);
			dprintf (DEBUG_TOKEN, "STRINGIT: Stored a string literal of length %d\n", stringlen);
		}

		if (save_sttype == 037)		/* If PURGE, no more processing. */
			break;
		
		if (*lp != ',')
			TOKERROR (31, "Comma expected")
		lp = get_next_character (lp + 1);

		sttype = 2;
		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (30, "Record length expression expected")
		stmt = nstmt;

		if (*lp != ',')
			break;
		lp = get_next_character (lp + 1);

		sttype = 2;
		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (30, "Missing record length")
		stmt = nstmt;

		break;
	case 040:	/* ADVANCE */
		lp = get_next_character (lp);
		stlen += 1;
		*stmt++ = ITOWH (sttype << 9);
		*stmt++ = ITOWL (sttype << 9);
		lp = get_next_character (lp);
		if (*lp != '#')
			TOKERROR (21, "File reference expected")
		lp = get_next_character(lp + 1);

		sttype = 4;
		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (34, "Bad or missing file reference")
		stmt = nstmt;
		lp = get_next_character (lp);


		if (*lp != ';')
			TOKERROR (33, "Semi-colon expected.")
		lp = get_next_character (lp + 1);

		sttype = 3;
		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (34, "Missing or bad numeric value")
		stmt = nstmt;
		lp = get_next_character(lp);
		
		if (*lp != ',')
			TOKERROR (39, "Comma expected")
		lp = get_next_character (lp + 1);
		sttype = 2;
		lp = get_simple_variable (lp, &varname, &vartype);
		if (!varname)
			TOKERROR (31, "Variable expected")
		stlen ++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		lp = get_next_character (lp);

		sttype = 0;
		nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		break;
	case 041:	/* UPDATE */
		stlen += 1;
		*stmt++ = ITOWH (sttype << 9);
		*stmt++ = ITOWL (sttype << 9);
		lp = get_next_character (lp);
		if (*lp != '#')
			TOKERROR (21, "'#' File reference expected")
		lp = get_next_character(lp + 1);

		sttype = 4;
		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (30, "File reference expected")
		stmt = nstmt;

		lp = get_next_character (lp);
		if (*lp != ';')
			TOKERROR (53, "Semi-colon expected")
		lp = get_next_character (lp + 1);
		sttype = 3;

		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
		if (stringlen != 9999) {
			lp = nlp;
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9));
			*stmt++ = ITOWL ((sttype << 9));
			sttype = 1;
			stlen ++;
			*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
			*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
			memcpy (stmt, litstring, stringlen);
			stlen += (stringlen + 1) / 2;
			stmt += stringlen + (stringlen % 2);
			dprintf (DEBUG_TOKEN, "STRINGIT: Stored a string literal of length %d\n", stringlen);
		} else if (lp != nlp) {
			lp = nlp;
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			dprintf(DEBUG_TOKEN, "STRINGIT: Stored a string variable.\n");

			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
		} else {
#ifdef OLD
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
			lp = nlp;
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
#endif

			stmt = nstmt;
			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (30, "String or numeric expression expected")
			stmt = nstmt;
		}


		break;
	case 042:	/* ASSIGN */
		lp = get_next_character (lp);
		if (*lp == '*') {
			stlen ++;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);

			sttype = 022;
			stlen ++;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			lp = get_next_character (lp + 1);

/*
			if (*lp != ',')
				TOKERROR (53, "Comma expected")
			lp = get_next_character (lp + 1);
			sttype = 2;
			stlen ++;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
*/

			if (*lp != ',')
				TOKERROR (53, "Comma expected")
			lp = get_next_character (lp + 1);
			sttype = 2;

			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (30, "File reference expected")
			stmt = nstmt;

			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

			if (*lp != ',')
				break;
			sttype = 2;
			lp = get_next_character (lp + 1);

			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
			lp = nlp;
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
		} else {
			nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
			dprintf (DEBUG_TOKEN, "get_source_string: RETURNED:  lp=%x  nlp=%x, varname=%d  vartype=%d,  litstring5='%-.5s' stringlen=%d\n",
					lp, nlp, varname, vartype, litstring, stringlen);
			lp = nlp;
			if (stringlen == 9999) {
				stlen ++;
				*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				dprintf(DEBUG_TOKEN, "STRINGIT: Stored a string variable.\n");

				sttype = 0;
				nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;
			} else {
				stlen ++;
				*stmt++ = ITOWH ((sttype << 9));
				*stmt++ = ITOWL ((sttype << 9));
				sttype = 1;
				stlen ++;
				*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
				*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
				memcpy (stmt, litstring, stringlen);
				stlen += (stringlen + 1) / 2;
				stmt += stringlen + (stringlen % 2);
				dprintf (DEBUG_TOKEN, "STRINGIT: Stored a string literal of length %d\n", stringlen);
			}

			if (*lp != ',')
				TOKERROR (53, "Comma expected")
			lp = get_next_character (lp + 1);
			sttype = 2;

			dprintf (DEBUG_TOKEN, "ASSIGN: Calling ans_formula with: %s\n", lp);
			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (30, "Missing file reference")
			stmt = nstmt;

			if (*lp != ',')
				TOKERROR (53, "Comma expected")
			lp = get_next_character (lp + 1);
			sttype = 2;

			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
			lp = nlp;
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

			/* COMMA -- FILE NUMBER -- RETURN VARIABLE -- [ COMMA PROTECTION MASK ] -- */
			/*	[ COMMA RESTRICTION] */

			if (*lp != ',') {
				break;
			}
			lp = get_next_character (lp + 1);
			sttype = 2;

			nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
			dprintf (DEBUG_TOKEN, "get_source_string: RETURNED:  lp=%x  nlp=%x, varname=%d  vartype=%d,  litstring5='%-.5s' stringlen=%d\n",
					lp, nlp, varname, vartype, litstring, stringlen);
			if (nlp != lp) {
				if (stringlen != 9999) {
					lp = nlp;
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9));
					*stmt++ = ITOWL ((sttype << 9));
					sttype = 1;
					stlen ++;
					*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
					*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);
					stlen += (stringlen + 1) / 2;
					stmt += stringlen + (stringlen % 2);
					dprintf (DEBUG_TOKEN, "STRINGIT: Stored a string literal of length %d\n", stringlen);
				} else if (stringlen == 9999 && nstmt != NULL) {
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
					*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
					dprintf(DEBUG_TOKEN, "STRINGIT: Stored a string variable.\n");

					sttype = 0;
					lp = nlp;
					nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
					if (nstmt == NULL)
						TOKERROR (30, "Bad or missing protection mask")
					stmt = nstmt;
				}
				lp = get_next_character (lp);
				if (*lp != ',')
					break;
				lp = get_next_character (lp + 1);
			}


			sttype = 2;
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);

			if ((nlp = match_literal (lp, "RR")) != NULL) {
				sttype = 044;
				stlen += 1;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				lp = nlp;
			} else if ((nlp = match_literal (lp, "WR")) != NULL) {
				sttype = 045;
				stlen += 1;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				lp = nlp;
			} else if ((nlp = match_literal (lp, "NR")) != NULL) {
				sttype = 046;
				stlen += 1;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				lp = nlp;
			} else
				TOKERROR (31, "Bad or missing restriction code")
			lp = get_next_character (lp);
			sttype = 0;
		}
		break;
	case 043:	/* LINPUT */
		lp = get_next_character (lp);
		if (*lp == '#') {
			lp = get_next_character (lp + 1);

			stlen ++;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);

			sttype = 4;

			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (30, "File reference expected")
			stmt = nstmt;

			if (*lp != ';')
				TOKERROR (55, "Semi-colon expected")
			lp = get_next_character (lp + 1);
			sttype = 3;
		}

		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
		}
		lp = nlp;

		stlen ++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		break;

	case 044:	/* IMAGE */
		if (*(lp - 1) == ' ')
			lp--;		/* Backup one character for IMAGE, I.E. "10 IMAGEABCD" is the formatting */
		stlen++;
		*stmt++ = ITOWH ((sttype << 9) | strlen(lp));
		*stmt++ = ITOWL ((sttype << 9) | strlen(lp));
		stlen += (strlen(lp) / 2) + (strlen(lp) % 2);
		memcpy(stmt, lp, strlen(lp));		/* String */
		stmt += strlen(lp) + (strlen(lp) % 2);
		if (strlen(lp) % 2)			/* If string spills over into */
			*(stmt++) = 0;			/*   next word, copy last byte */
		lp += strlen(lp);
		break;
	case 045:	/* COM */
	case 047:	/* DIM */
		save_sttype = sttype;
		lp = get_next_character (lp);
		while (*lp) {
			dim1 = 0;
			dim2 = 0;
			dimvar = 0;
			nlp = get_string_variable (lp, &varname, &vartype);
			if (lp == nlp) {
					if (isalpha (*lp) && (*get_nth_character (lp, 1) == '(' || *get_nth_character (lp,1) == '[')) {
					dimvar = *lp;
					lp = get_next_character (lp + 1);
	
					if (*lp != '(' && *lp != '[')
						TOKERROR (61, "Missing left parenthesis")
					lp = get_next_character (lp + 1);
	
					nlp = get_number (lp, GN_INT, &dim1);
					if (nlp == NULL)
						TOKERROR (62, "Dimension expected")
					if (dim1 < 1 || dim1 > 9999)
						TOKERROR (80, "Illegal subscript")
					lp = get_next_character(nlp);
	
					if (*lp == ',') {
						lp = get_next_character (lp + 1);
						nlp = get_number (lp, GN_INT, &dim2);
						if (nlp == NULL)
							TOKERROR (64, "Invalid second dimension")
						if (dim2 < 1 || dim2 > 9999)
							TOKERROR (65, "Dimension out of range")
						lp = nlp;
					}
	
					if (*lp != ')' && *lp != ']')
						TOKERROR (66, "Missing dimension right parenthesis")
					lp = get_next_character (lp + 1);
	
					dimvar = dimvar - '@';
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9) | ((dimvar & 037) << 4) | ((dim2 > 0) ? 2 : 1));
					*stmt++ = ITOWL ((sttype << 9) | ((dimvar & 037) << 4) | ((dim2 > 0) ? 2 : 1));
					sttype = 012;	/* [ */
					stlen ++;
					*stmt++ = ITOWH (0x8000 | (sttype << 9) | 3);
					*stmt++ = ITOWL (0x8000 | (sttype << 9) | 3);
					stlen ++;
					*stmt++ = ITOWH (dim1 & 0xFFFF);
					*stmt++ = ITOWL (dim1 & 0xFFFF);
					if (dim2 > 0) {
						sttype = 002;	/* , */
						stlen ++;
						*stmt++ = ITOWH (0x8000 | (sttype << 9) | 3);
						*stmt++ = ITOWL (0x8000 | (sttype << 9) | 3);
						stlen ++;
						*stmt++ = ITOWH (dim2 & 0xFFFF);
						*stmt++ = ITOWL (dim2 & 0xFFFF);
					}
					sttype = 011;	/* ] */
					stlen ++;
					*stmt++ = ITOWH (sttype << 9);
					*stmt++ = ITOWL (sttype << 9);
				} else if (save_sttype == 045) {  /* Allow regular variables for COM */
					nlp = get_simple_variable (lp, &varname, &vartype);
					if (!varname)
						TOKERROR (73, "Invalid variable")
					lp = nlp;
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | vartype);
					*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | vartype);
				} else
					TOKERROR (60, "Invalid subscripted variable")
			} else {
				lp = get_next_character (nlp);
	
				if (*lp != '(' && *lp != '[')
					TOKERROR (67, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
	
				nlp = get_number (lp, GN_INT, &dim1);
				if ( nlp == NULL || dim1 < 1 || dim1 > 9999)
					TOKERROR (68, "Illegal string dimension")
				lp = get_next_character(nlp);
	
				if (*lp != ')' && *lp != ']')
					TOKERROR (70, "Right parenthesis expected")
				lp = get_next_character (lp + 1);
				
				stlen++;
				*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | vartype);
				*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | vartype);
				sttype = 012;   /* [ */
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9) | 3);
				*stmt++ = ITOWL (0x8000 | (sttype << 9) | 3);
				stlen ++;
				*stmt++ = ITOWH (dim1 & 0xFFFF);
				*stmt++ = ITOWL (dim1 & 0xFFFF);
				sttype = 011;   /* ] */
				stlen ++;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
			}
			if (*lp && *lp != ',')
				TOKERROR (71, "Comma expected")
			if (*lp == ',') {
				lp = get_next_character (lp + 1);
				if (!*lp)
					TOKERROR (72, "Extraneous comma")
				sttype = 002;
			}
		}
		break;

	case 050:	/* DEF */
		lp = get_next_character (lp);
		/* Check for FNx */
		if (*lp != 'F' || *(get_next_character(lp + 1)) != 'N')
			TOKERROR (4, "FNx function name expected")
		vartype=017;
		lp = get_next_character (lp + 1);
		lp = get_next_character (lp + 1);

		if (!isalpha(*lp))
			TOKERROR (4, "Function name not FNA through FNZ")
		varname = toupper(*lp) - '@';
		stlen ++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		lp = get_next_character (lp + 1);

		if (*lp != '(' && *lp != '[')
			TOKERROR (7, "Left parenthensis expected")
		sttype = 013;	/* Left Paren */
		lp = get_next_character (lp + 1);

		vartype = 4;
		if (!isalpha(*lp))
			TOKERROR (5, "Invalid user-defined function name")
		varname = toupper(*lp) - '@';
		lp = get_next_character (lp + 1);

		if (isdigit(*lp)) {
			vartype = 5 + (*lp - '0');
			dprintf (DEBUG_TOKEN, "VARTYPE = %d\n", vartype);
			lp = get_next_character (lp + 1);
		}

		stlen ++;
		*stmt++ = ITOWH (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		if (*lp != ')' && *lp != ']')
			TOKERROR (7, "Right parenthesis expected")

		sttype = 010;	/* Right Paren */
		varname = 0;
		vartype = 0;
		stlen++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		lp = get_next_character (lp + 1);
		if (*lp != '=')
			TOKERROR (6, "Assignment operator expected")
		lp = get_next_character (lp + 1);

		sttype = 017;	/* '=' assignment */
		varname = 0;
		vartype = 0;
		varnum = 99;
		
		dprintf (DEBUG_TOKEN, "DEFFN: IN stlen=%d\n", stlen);
		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		dprintf (DEBUG_TOKEN, "DEFFN: OUT stlen=%d\n", stlen);
		if (stmt == nstmt)
			TOKERROR (35, "Invalid function defininition")
		stmt = nstmt;

		dprintf (DEBUG_TOKEN, "DEF FN function passed parsing.\n");
		break;

	case 051:	/* REM */
	case 070:	/* FILES */

		stlen ++;
		*stmt++ = ITOWH (sttype << 9 | *lp);
		*stmt++ = ITOWL (sttype << 9 | *lp);
		lp++;

		stlen += (strlen(lp) + 1) / 2;
		memcpy(stmt, lp, strlen(lp));		/* String */
		stmt += strlen(lp);

		if ((strlen(lp) % 2))
			*stmt++ = 0;

		lp = lp + strlen(lp);
		break;

	case 052:	/* GOTO */
	case 056:	/* GOSUB */
		lp = get_next_character (lp);
		nlp = get_number (lp, GN_LINENO, &wlineno);
		if (nlp != NULL && *(get_next_character(nlp)) == 0) {
			dprintf (DEBUG_TOKEN, "GOTO: Got a line number: %d\n", wlineno);
			lp = nlp;
			/* We got a simple GOTO lineno */
			dprintf (DEBUG_TOKEN, "GOTO: SIMPLE GOTO FOUND\n");
			stlen += 1;
			*stmt++ = ITOWH (0x8000 | (sttype << 9) | 3);
			*stmt++ = ITOWL (0x8000 | (sttype << 9) | 3);
			stlen += 1;
			ilineno = wlineno;
			*stmt++ = ITOWH (ilineno);
			*stmt++ = ITOWL (ilineno);
			sttype = 0;
			varname = 0;
			vartype = 0;
			varnum = 99;
		} else {
			varname = 0;
			vartype = 0;
			varnum = 99;
			dprintf (DEBUG_TOKEN, "GOTOans: IN stlen=%d  stmt=0x%x\n", stlen, stmt);
			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			dprintf (DEBUG_TOKEN, "GOTOans: OUT stlen=%d stmt=0x%x\n", stlen, stmt);
			if (stmt == nstmt)
				TOKERROR (19, "Invalid GOTO syntax")
			stmt = nstmt;
			if (*(get_next_character(lp)) != 'O' || 
					*(get_nth_character(lp, 1)) != 'F')
				TOKERROR (20, "Invalid GOTO ... OF syntax")
			lp = get_next_character (lp + 1);
			lp = get_next_character (lp + 1);
			
			sttype = 074;  /* OF */
			stlen += 1;
			*stmt++ = ITOWH (0x8000 | (sttype << 9) | 3);
			*stmt++ = ITOWL (0x8000 | (sttype << 9) | 3);

			dprintf (DEBUG_TOKEN, "GOTO: Checking for statement numbers -- next char lp=%d '%c'\n", *lp, *lp);
			lp = get_next_character(lp);
			while ((lp = get_number (lp, GN_LINENO, &wlineno)) != NULL) {
				dprintf (DEBUG_TOKEN, "GOTO: STORING statement number: %d\n", wlineno);
				stlen += 1;
				ilineno = wlineno;
				*stmt++ = ITOWH (ilineno);
				*stmt++ = ITOWL (ilineno);
				if (*lp == 0)
					break;
				if (*lp != ',')
					TOKERROR (21, "Comma expected")
				lp = get_next_character(lp + 1);
			}
			
			dprintf (DEBUG_TOKEN, "GOTOans: GOT GOTO... OF\n");
			dprintf (DEBUG_TOKEN, "GOTOans: OUT stlen=%d\n", stlen);
		}

		break;

	case 053:	/* IF */
		lp = get_next_character (lp);
		varnum = 99;
		vartype = 0;
		varname = 0;

		if ((nlp = match_literal (lp, "END")) != NULL) {
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			/* END keyword found */
			lp = nlp;
			if (*lp != '#')
				TOKERROR (21, "'#' file reference expected")
			lp = get_next_character(lp + 1);
			sttype = 060;
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			sttype = 4;
			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (37, "Invalid file reference")
			stmt = nstmt;
		} else if ((nlp = match_literal (lp, "ERROR")) != NULL) {
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			/* ERROR keyword found */
			sttype = 047;
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			lp = nlp;
		} else if ((nlp = get_string_variable (lp, &varname, &vartype)) != lp) {
			lp = nlp;
			stlen++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			sttype = 0;

			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

			sttype = 0;
			varname = 0;
			vartype = 0;

			dprintf (DEBUG_TOKEN, "IF:::::::: Getting relational operator\n");
			nlp = get_relational_operator (lp, &sttype);
			if (lp == nlp)
				TOKERROR (57, "Relational operator expected")
			lp = nlp;
			dprintf (DEBUG_TOKEN, "IF;;;;;;;; Got relational operator, *lp = %c\n", *lp);

			if ((nlp = match_literal (lp, "CHR$")) != NULL) {
				lp = nlp;

				varname = 027;
				vartype = 017;
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
				if (stmt == nstmt)
					TOKERROR (34, "Invalid or missing numeric argument to CHR$");
				stmt = nstmt;

				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				stlen ++;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				sttype = 0;

			} else {

				nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
				dprintf (DEBUG_TOKEN, "get_source_string: RETURNED:  lp=%x  nlp=%x, varname=%d  vartype=%d,  litstring5='%-.5s' stringlen=%d *lp=%c\n",
						lp, nlp, varname, vartype, litstring, stringlen, *lp);
				lp = nlp;
				if (stringlen == 9999) {
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
					*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

					sttype = 0;
					nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;
					dprintf(DEBUG_TOKEN, "STRINGIT: Stored a string variable, *lp=%c.\n", *lp);
				} else {
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9));
					*stmt++ = ITOWL ((sttype << 9));
					sttype = 1;
					stlen ++;
					*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
					*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);
					stlen += (stringlen + 1) / 2;
					stmt += stringlen + (stringlen % 2);
					dprintf (DEBUG_TOKEN, "STRINGIT: Stored a string literal of length %d\n", stringlen);
				}
			}

		} else {
			dprintf (DEBUG_TOKEN, "IF: Calling ans_formula: lp=%s\n", lp);
			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (38, "Numeric expression expected")
			stmt = nstmt;
		}
			
		/* Demand THEN */
		if ((nlp = match_literal (lp, "THEN")) == NULL)
			TOKERROR (22, "THEN expected")
		sttype = 075;
		stlen += 1;
		*stmt++ = ITOWH (sttype << 9);
		*stmt++ = ITOWL (sttype << 9);
		lp = get_next_character (nlp);

		/* Demand a statement number now */
		if ((lp = get_number (lp, GN_LINENO, &wlineno)) == NULL)
			TOKERROR (23, "Statement number expected")
		stlen += 1;
		*stmt++ = ITOWH (0x8000 | 3);
		*stmt++ = ITOWL (0x8000 | 3);
		ilineno = wlineno;
		stlen += 1;
		*stmt++ = ITOWH (ilineno);
		*stmt++ = ITOWL (ilineno);
		break;

	case 054:	/* FOR */
		lp = get_next_character (lp);
		/* Demand a simple variable */
		nlp = get_simple_variable (lp, &varname, &vartype);
		if (lp == nlp)
			TOKERROR (42, "Numeric variable expected")
		lp = get_next_character (nlp);
		stlen++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		varname = 0;
		vartype = 0;

		/* Demand an assignment '=' */
		if (*lp != '=')
			TOKERROR (56, "'=' expected")
		lp = get_next_character (lp + 1);
		sttype = 030;

		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (38, "Numeric expression expected")
		stmt = nstmt;

		/* Demand TO */
		if ((nlp = match_literal (lp, "TO")) == NULL)
			TOKERROR (57, "TO expected")
		lp = get_next_character(nlp);
		sttype = 076;

		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (38, "Numeric expression expected")
		stmt = nstmt;

		/* Optional STEP */
		lp = get_next_character (lp);
		if ((nlp = match_literal (lp, "STEP")) == NULL)
			break;
		lp = nlp;
		sttype = 077;

		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (38, "Numeric expression expected")
		stmt = nstmt;

		break;

	case 055:	/* NEXT */
		lp = get_next_character (lp);
		lp = get_simple_variable (lp, &varname, &vartype);
		if (varname == 0)
			TOKERROR (27, "Numeric variable expected")
		stlen++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		break;
	case 057:	/* RETURN */
	case 060:	/* END */
	case 061:	/* STOP */
		stlen += 1;
		*stmt++ = ITOWH (sttype << 9);
		*stmt++ = ITOWL (sttype << 9);
		break;
	case 062:	/* DATA */
		lp = get_next_character (lp);
		while (*lp) {
			nlp = get_string_literal (lp, litstring, &stringlen);
			if (lp == nlp) {
				nlp = get_number (lp, GN_FLOAT, &wnumber);
				if (nlp == NULL)
					TOKERROR (58, "Invalid numeric or string expression for DATA")
				lp = nlp;
				dprintf (DEBUG_TOKEN, "DATA: Storing a number: %G\n", wnumber);
				HPfloat = IEEEtoHPf(wnumber);
				dprintf (DEBUG_TOKEN, "DATA: Checking wnumber=%G conversion: %G\n", wnumber, HPftoIEEE(HPfloat));
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9));
				*stmt++ = ITOWL (0x8000 | (sttype << 9));
				stlen++;
				*stmt++ = ITOWH (HPfloat >> 16);
				*stmt++ = ITOWL (HPfloat >> 16);
				stlen++;
				*stmt++ = ITOWH (HPfloat & 0xFFFF);
				*stmt++ = ITOWL (HPfloat & 0xFFFF);
			} else {
				lp = nlp;
				stlen ++;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				stlen ++;
				sttype = 1;
				*stmt++ = ITOWH ((sttype << 9) | (stringlen & 0377));
				*stmt++ = ITOWL ((sttype << 9) | (stringlen & 0377));
				memcpy (stmt, litstring, stringlen);
				stlen += (stringlen + 1) / 2;
				stmt += stringlen + (stringlen % 2);
			}
			sttype = 0;
			lp = get_next_character (lp);
			if (*lp != ',')
				break;
			sttype = 2;
			lp = get_next_character (lp + 1);
		}
		
		break;
	case 063:	/* INPUT */
		lp = get_next_character (lp);
		while (*lp) {
			nlp = get_string_variable (lp, &varname, &vartype);
			if (lp != nlp) {
				lp = nlp;
			} else {
				nlp = get_simple_variable (lp, &varname, &vartype);
				if (lp == nlp)
					TOKERROR (34, "Variable expected")
				lp = nlp;
			}
			stlen++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4)
					| (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4)
					| (vartype & 017));

			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

			if (*lp != ',')
				break;
			lp = get_next_character (lp + 1);
			sttype = 002;
		}
		break;
	case 064:	/* READ */
		lp = get_next_character (lp);
		fileref = 0;
		if (*lp == '#') {
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			lp = get_next_character (lp + 1);

			sttype = 4;
			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (34, "File reference expected")
			stmt = nstmt;

			if (*lp) {
				if (*lp == ',') {
					lp = get_next_character (lp + 1);
					sttype = 2;
					nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
					if (stmt == nstmt)
						TOKERROR (34, "Record number expected")
					stmt = nstmt;
				} else if (*lp != ';')
					TOKERROR (74, "Semi-colon expected")
				else {
					sttype = 3;
					lp = get_next_character (lp + 1);
				}
				fileref = 1;
			} else
				break;		/* Allow simple "READ #n" statements. */
		}

		while (*lp) {
			nlp = get_string_variable (lp, &varname, &vartype);
			if (lp == nlp) {
				nlp = get_simple_variable (lp, &varname, &vartype);
				if (lp == nlp)
					TOKERROR (42, "Variable expected")
			}
			lp = nlp;
			dprintf (DEBUG_TOKEN, "READ: Got variable, storing varname=%d  vartype=%o  sttype=%o   next pos=%c\n",
					varname, vartype, sttype, *lp);
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));


				sttype = 0;
				nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;
			sttype = 0;

			if (*lp) {
				if (*lp != ',' && *lp != ';')
					TOKERROR (50, "List delimiter expected")
				lp = get_next_character (lp + 1);
			} else
				break;
			sttype = 002;
		}
		break;
	case 065:	/* PRINT */
		lp = get_next_character (lp);

		save_sttype = sttype;

		if (*lp == '#' ) {
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);

			lp = get_next_character (lp + 1);

			sttype = 4;
			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (34, "File reference expected")
			stmt = nstmt;

			if (*lp && *lp == ',') {
				lp = get_next_character (lp + 1);
				sttype = 2;
				nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
				if (stmt == nstmt)
					TOKERROR (34, "Record number expression expected")
				stmt = nstmt;
			}

			if (*lp) {
				if (*lp != ';')
					TOKERROR (91, "Semi-colon expected")
				sttype = 3;
				lp = get_next_character (lp + 1);
			}
		}

		if ((nlp = match_literal (lp, "USING")) != NULL) {
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			sttype = 043;

			lp = nlp;

			/* TODO: Get a FORMAT-STRING or a STATEMENT NUMBER */
			wlineno = 0;
			nlp = get_number (lp, GN_LINENO, &wlineno);
			if (nlp != NULL) {
				/* Statement number */
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9) | 3);
				*stmt++ = ITOWL (0x8000 | (sttype << 9) | 3);
				lp = nlp;
				stlen += 1;
				ilineno = wlineno;
				*stmt++ = ITOWH (ilineno);
				*stmt++ = ITOWL (ilineno);
			} else {
				/* TODO: Get a FORMAT-STRING */
				nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
				if (stringlen == 9999) {
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
					*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

					sttype = 0;
					lp = nlp;

					nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;

					dprintf(DEBUG_TOKEN, "STRINGIT: Stored a string variable, *lp=%c.\n", *lp);
				} else {
					lp = nlp;
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9));
					*stmt++ = ITOWL ((sttype << 9));
					sttype = 1;
					stlen ++;
					*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
					*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);
					stlen += (stringlen + 1) / 2;
					stmt += stringlen;
					if (stringlen % 2) {
						*stmt++ = 0;
					}
					dprintf (DEBUG_TOKEN, "STRINGIT: Stored a string literal of length %d\n", stringlen);
				}
				sttype = 0;
			}

			if (*lp) {
				if (*lp != ';')
					TOKERROR (91, "Semi-colon expected")
				sttype = 3;
				lp = get_next_character (lp + 1);
			} else
				break;
		}

		while (*lp) {
			if ((nlp = match_literal (lp, "UPS$")) != NULL) {
				lp = nlp;

				varname = 030;
				vartype = 017;
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
				if (stringlen != 9999) {
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9));
					*stmt++ = ITOWL ((sttype << 9));
					sttype = 1;
					stlen ++;
					*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
					*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);
					stlen += (stringlen + 1) / 2;
					stmt += stringlen + (stringlen % 2);
					dprintf (DEBUG_TOKEN, "STRINGIT2: Stored a string literal of length %d\n", stringlen);
					lp = nlp;
				} else if (nlp != lp) {
					lp = nlp;
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
					*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
					dprintf(DEBUG_TOKEN, "STRINGIT2: Stored a string variable, varname=%d  vartype=%d, stringlen=%d.\n", varname, vartype, stringlen);
					sttype = 0;
					nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;
				}
				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				stlen ++;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				sttype = 0;

			}  else if ((nlp = match_literal (lp, "CHR$")) != NULL) {
				lp = nlp;

				varname = 027;
				vartype = 017;
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
				if (stmt == nstmt)
				TOKERROR (34, "Expecting numeric argument to CHR$")
				stmt = nstmt;

				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				stlen ++;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				sttype = 0;

			}  else if ((nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen)) != lp &&
					 			stringlen != 9999) {
				lp = nlp;
				if (sttype) {
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9));
					*stmt++ = ITOWL ((sttype << 9));
				}

				dprintf (DEBUG_TOKEN, "STRINGIT2: Storing a string literal of length %d, string=\"%s\", sttype=%o stlen=%d\n", stringlen, litstring, sttype, stlen);
				sttype = 1;
				fixstringlen = stringlen;

				stlen ++;
				*stmt++ = ITOWH ((sttype << 9) | (fixstringlen & 0377));
				*stmt++ = ITOWL ((sttype << 9) | (fixstringlen & 0377));

				memcpy (stmt, litstring, stringlen);
				stlen = stlen + (stringlen + 1) / 2;
				stmt += stringlen;
				if (stringlen % 2) {
					*stmt++ = 0;
				}

				dprintf (DEBUG_TOKEN, "STRINGIT2: Stored a string literal of length %d, string=\"%s\", sttype=%o stlen=%d\n", stringlen, litstring, sttype, stlen);
				sttype = 0;
			} else if (nlp != lp) {
				lp = nlp;
				stlen ++;
				*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				dprintf(DEBUG_TOKEN, "STRINGIT2: Stored a string variable, varname=%d  vartype=%d, stringlen=%d.\n", varname, vartype, stringlen);
/*eprintf("STRINGIT2: Stored a string variable, varname=%d(%c)  vartype=%d, stringlen=%d.\n", varname, varname + 64, vartype, stringlen);*/
				sttype = 0;
				varname = 0;
				vartype = 0;
				nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;
			} else if (save_sttype == 065 && (nlp = match_literal (lp, "END")) != NULL) {
				stlen += 1;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				sttype = 060;
				lp = nlp;
				break;
			} else {
				nlp = match_token (lp, &new_sttype, Bfunc, Bfunc_count, 0);
				if (nlp != lp && new_sttype <= 3) {
					/*sttype = new_sttype;*/
					stlen += 1;
					*stmt++ = ITOWH (0x8000 | (sttype << 9) | (new_sttype << 4) | 017);
					*stmt++ = ITOWL (0x8000 | (sttype << 9) | (new_sttype << 4) | 017);
					sttype = 0;
					lp = nlp;

					if (*lp != '(' && *lp != '[')
						TOKERROR (93, "Left parenthesis expected")

					sttype = 013;
					lp = get_next_character (lp + 1);

					nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
					if (stmt == nstmt)
						TOKERROR (95, "Numeric expression expected")
					stmt = nstmt;

					if (*lp != ')' && *lp != ']')
						TOKERROR (96, "Right parentheses expected")

					sttype = 010;
					stlen += 1;
					*stmt++ = ITOWH (sttype << 9);
					*stmt++ = ITOWL (sttype << 9);
					sttype = 0;
					lp = get_next_character (lp + 1);
					
				} else if ((nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum)) != stmt) {
					stmt = nstmt;
					sttype = 0;
				} else {
					nlp = get_simple_variable (lp, &varname, &vartype);
					if (lp != nlp) {
						lp = nlp;
						stlen++;
						*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
						*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
						varname = 0;
						vartype = 0;
	
						sttype = 0;
						nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
						if (nstmt == NULL)
							TOKERROR (80, "Invalid subscript")
						stmt = nstmt;
					} else
						TOKERROR (98, "Invalid syntax")
				}
			}

			if (*lp == ';') {
				lp = get_next_character (lp + 1);
				if (sttype) {
					stlen += 1;
					*stmt++ = ITOWH (sttype << 9);
					*stmt++ = ITOWL (sttype << 9);
				}
				sttype = 3;
			} else if (*lp == ',') {
				lp = get_next_character (lp + 1);
				if (sttype) {
					stlen += 1;
					*stmt++ = ITOWH (sttype << 9);
					*stmt++ = ITOWL (sttype << 9);
				}
				sttype = 2;
			} else if (!*lp)
				break;
			else
				sttype = 0;	/* List delimited are not mandated in PRINT */
		}

		if (sttype) {
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
		}

		break;
	case 066:	/* RESTORE */
		lp = get_next_character (lp);
		wlineno = 0;
		nlp = get_number (lp, GN_LINENO, &wlineno);
		stlen += 1;
		if (nlp == NULL) {
			if (*lp)
				TOKERROR (29, "Statement number expected")
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
		} else {
			/* Optional statement number */
			*stmt++ = ITOWH (0x8000 | (sttype << 9) | 3);
			*stmt++ = ITOWL (0x8000 | (sttype << 9) | 3);
			lp = nlp;
			stlen += 1;
			ilineno = wlineno;
			*stmt++ = ITOWH (ilineno);
			*stmt++ = ITOWL (ilineno);
		}
		break;
	case 067:	/* MAT */
		/* TODO */
		/*
		 * MAT PRINT				<array-variable> | <print-function> [ ,  ; ... ]
		 * MAT PRINT [ #<numeric-expression ;]	[ <array-variable> | <print-function> ] [ ,  ; ... ] [END]
		 * MAT PRINT USING  <format-expression>	<array-variable> | <print-function> [ ,  ; ... ]
		 * MAT READ [ #<numeric-expression> ;]	<array-variable> [ ( <dimensions ) ] [ ... ]
		 * MAT INPUT 				<array-variable> [ ( <dimensions ) ] [ ... ]
		 * MAT  				<array-variable> = { ZER | CON | IDN }
		 * MAT					<array-variable> = <array-variable> [ +|-|* <array-variable> ]
		 * MAT					<array-variable> = {TRN | INV} ( <array-variable> )
		 */

		
		if ((nlp = match_literal (lp, "PRINT")) != NULL) {
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			sttype = 065;
			lp = nlp;
			if ((nlp = match_literal (lp, "USING")) != NULL) {
				stlen += 1;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				sttype = 043;
				lp = nlp;

				/* TODO: Get a FORMAT-STRING or a STATEMENT NUMBER */
				wlineno = 0;
				nlp = get_number (lp, GN_LINENO, &wlineno);
				if (nlp != NULL) {
					/* Statement number */
					stlen ++;
					*stmt++ = ITOWH (0x8000 | (sttype << 9) | 3);
					*stmt++ = ITOWL (0x8000 | (sttype << 9) | 3);
					lp = nlp;
					stlen += 1;
					ilineno = wlineno;
					*stmt++ = ITOWH (ilineno);
					*stmt++ = ITOWL (ilineno);
				} else {
					/* TODO: Get a FORMAT-STRING */
					nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
					if (stringlen == 9999) {
						stlen ++;
						*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
						*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
	
						sttype = 0;
						lp = nlp;
	
						nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
						if (nstmt == NULL)
							TOKERROR (80, "Invalid subscript")
						stmt = nstmt;
	
					} else {
						lp = nlp;
						stlen ++;
						*stmt++ = ITOWH ((sttype << 9));
						*stmt++ = ITOWL ((sttype << 9));
						sttype = 1;
						stlen ++;
						*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
						*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
						memcpy (stmt, litstring, stringlen);
						stlen += (stringlen + 1) / 2;
						stmt += stringlen;
						if (stringlen % 2) {
							*stmt++ = 0;
						}
					}
					sttype = 0;
				}
				if (*lp) {
					if (*lp != ';')
						TOKERROR (91, "Semi-colon expected")
					sttype = 3;
					lp = get_next_character (lp + 1);
				} else
					break;
			}
		} else if ((nlp = match_literal (lp, "READ")) != NULL) {
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			sttype = 064;
			lp = nlp;
		} else if ((nlp = match_literal (lp, "INPUT")) != NULL) {
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			sttype = 063;
			lp = nlp;
		}

		save_sttype = sttype;	/* Keep track of what MAT format we're parsing. */
		new_sttype = 0;		/* Keep track of whether we're reading or writing a file. */

		if ((sttype == 065 || sttype == 064) && *lp == '#') {
			new_sttype = 1;
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
			sttype= 4;
			lp = get_next_character (lp + 1);

			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (34, "Missing or invalid file reference")
			stmt = nstmt;

			if (*lp == ',') {
				lp = get_next_character (lp + 1);
				sttype = 2;

				nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
				if (stmt == nstmt)
					TOKERROR (34, "Missing or invalid record number")
				stmt = nstmt;
			}

			if (*lp != ';')
				TOKERROR (38, "Semi-colon expected")
			lp = get_next_character (lp + 1);
			if (sttype)
				TOKERROR (-99, "INTERNAL ERROR #MAT1: left-over sttype undealt with")
			sttype= 3;
		}
			
		if (save_sttype == 0 && new_sttype == 1) {
			/* allow END or a print-function for 'MAT PRINT #' */
			/* Just fall through for a print-function, but process and finish for END */
			/* otherwise error */
		}

		if (save_sttype == 065 && new_sttype == 1 && !*lp)
			break;

		nlp = get_simple_variable (lp, &varname, &vartype);
		if (lp != nlp) {

			lp = nlp;
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | vartype);
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | vartype);

			sttype = 0;
			varname = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
		} else
			TOKERROR (37, "Array variable expected")


		if (save_sttype == 067) {	/* Plain old MAT =? */
			if (!*lp)
				break;
			if (*lp != '=')
				TOKERROR (36, "'=' assignment operator expected")
			sttype = 017;		/* Assignment '=' */
			lp = get_next_character (lp + 1);

			if ((nlp = match_token (lp, &new_sttype, Bfunc, Bfunc_count, 5)) != lp) {
				/* Got a MAT assignment function */

				if (new_sttype >= 033 && new_sttype <= 035) {

					lp = nlp;
					stlen ++;
					*stmt++ = ITOWH (0x8000 | (sttype << 9) | (new_sttype << 4) | 017);
					*stmt++ = ITOWL (0x8000 | (sttype << 9) | (new_sttype << 4) | 017);

					sttype = 0;
					varname = 0;
					nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;
					break;
				} else if (new_sttype < 033)
					TOKERROR (81, "Invalid function for MAT")

				lp = nlp;
				if (*lp != '[' && *lp != '(')
					TOKERROR (81, "Left parenthesis expected")

				lp = nlp;
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9) | (new_sttype << 4) | 017);
				*stmt++ = ITOWL (0x8000 | (sttype << 9) | (new_sttype << 4) | 017);

				sttype = 0;
				varname = 0;
				nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
				if (stmt == nstmt) {
					eprintf ("ERROR: Missing numeric expression after program name.\n");
				}
				stmt = nstmt;
				break;

/*WORK*/

/*
				lp = nlp;
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9) | (varname << 4) | 017);
				*stmt++ = ITOWL (0x8000 | (sttype << 9) | (varname << 4) | 017);
				sttype = 0;
				varname = 0;
*/
			}
			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt) {
				eprintf ("ERROR: Missing numeric expression after program name.\n");
			}
			stmt = nstmt;

		}

		
		while (*lp) {
			if (!*lp)
				break;
			if (*lp == ',') {
				sttype = 2;
			} else if (*lp == ';') {
				sttype = 3;
			} else
				TOKERROR (50, "List delimiter expected")
			lp = get_next_character (lp + 1);
			if (!*lp)
				break;

			if (save_sttype == 065 && new_sttype == 1 && (nlp = match_literal (lp, "END")) != NULL) {
				stlen += 1;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				sttype = 060;
				lp = nlp;
				break;
			}
				
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp != nlp) {
				lp = nlp;
				stlen ++;
				*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | vartype);
				*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | vartype);

				sttype = 0;
				varname = 0;
				nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;
			} else
				TOKERROR (84, "Invalid MAT PRINT source")
		}
		if (sttype) {
			stlen ++;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);
		}
		
		break;

	case 071:	/* CHAIN */
		/* TODO: [<return-variable> , ] <program-name> [, <numeric-expression> ] */

		/* Dummy call to get_source_string to see if a string follows */
		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);

		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp != nlp) {
				lp = nlp;
				stlen ++;
				*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

				sttype = 0;
				nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;

				if (*lp != ',')
					TOKERROR (50, "Comma expected")
				lp = get_next_character (lp + 1);
				sttype = 2;
			}
		}

		nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
		if (stringlen != 9999) {
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9));
			*stmt++ = ITOWL ((sttype << 9));
			sttype = 1;
			stlen ++;
			*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
			*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
			memcpy (stmt, litstring, stringlen);
			stlen += (stringlen + 1) / 2;
			stmt += stringlen + (stringlen % 2);
			dprintf (DEBUG_TOKEN, "STRINGIT2: Stored a string literal of length %d\n", stringlen);
			lp = nlp;
		} else if (varname != 0) {
			lp = nlp;
			stlen ++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;

		} else
			TOKERROR (49, "String expression expected for program name")

		lp = get_next_character (lp);
		if (!*lp)
			break;
		if (*lp != ',')
			TOKERROR (50, "Comma expected")
		lp = get_next_character (lp + 1);
		sttype = 2;

		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt) {
			eprintf ("ERROR: Missing numeric expression after program name.\n");
		}
		stmt = nstmt;
		break;

	case 072:	/* ENTER */
		lp = get_next_character (lp);

		if (*lp == '#') {
			lp = get_next_character (lp + 1);
			/* Store the ENTER statement code */
			stlen += 1;
			*stmt++ = ITOWH (sttype << 9);
			*stmt++ = ITOWL (sttype << 9);

			sttype = 4;
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (39, "Numeric variable expected")
			lp = nlp;
			stlen++;
			*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
			*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

			sttype = 0;
			nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
			if (nstmt == NULL)
				TOKERROR (80, "Invalid subscript")
			stmt = nstmt;
			if (*lp == ',')
				sttype = 2;
			else {
				if (!*lp)
					break;
				TOKERROR (39, "Comma expected")
			}
			lp = get_next_character (lp + 1);
		}
			
		nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
		if (stmt == nstmt)
			TOKERROR (34, "Bad or missing Time Allowed expression")
		stmt = nstmt;

		if (*lp != ',')
			TOKERROR (39, "Comma expected")
		lp = get_next_character (lp + 1);
		sttype = 2;

		nlp = get_simple_variable (lp, &varname, &vartype);
		if (lp == nlp)
			TOKERROR (34, "Numeric variable expected")
		lp = nlp;

		stlen++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		if (*lp != ',')
			TOKERROR (39, "Comma expected")
		lp = get_next_character (lp + 1);
		sttype = 2;

		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected for returned value")
		}
		lp = nlp;
		stlen ++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		sttype = 0;
		nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		break;
	case 046:	/* LET */
	case 073:	/* Implied LET */
		lp = get_next_character (lp);
		if (!isalpha(*lp))
			TOKERROR (3, "Unknown statement type, assignment operator expected")
		varnum = 99;
		vartype = 0;
		varname = 0;

		nlp = get_string_variable (lp, &varname, &vartype);
		if (lp == nlp) {
			nlp = get_simple_variable (lp, &varname, &vartype);
			if (lp == nlp)
				TOKERROR (42, "Variable expected")
		}
		lp = nlp;
		stlen ++;
		*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
		*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

		sttype = 0;
		nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
		if (nstmt == NULL)
			TOKERROR (80, "Invalid subscript")
		stmt = nstmt;

		lp = get_next_character (lp);
		if (*lp != '=')
			TOKERROR (-10, "Unknown statment, assignment operator expected")
		lp = get_next_character (lp + 1);
		sttype = 030;		/* Type is now '=' (assignment) */

		if (vartype == 0 || (varname >= 033 && varname <= 036)) {
		
			nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
			dprintf (DEBUG_TOKEN, "get_source_string: RETURNED:  lp=%x  nlp=%x, varname=%d  vartype=%d,  litstring5='%-.5s' stringlen=%d\n",
					lp, nlp, varname, vartype, litstring, stringlen);
			if (stringlen != 9999) {
				stlen ++;
				*stmt++ = ITOWH ((sttype << 9));
				*stmt++ = ITOWL ((sttype << 9));
				sttype = 1;
				stlen ++;
				*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
				*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
				memcpy (stmt, litstring, stringlen);
				stlen += (stringlen + 1) / 2;
				stmt += stringlen + (stringlen % 2);
				dprintf (DEBUG_TOKEN, "STRINGIT2: Stored a string literal of length %d\n", stringlen);
				lp = nlp;
			} else if (lp != nlp) {
				lp = nlp;
				stlen ++;
				*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				dprintf(DEBUG_TOKEN, "STRINGIT2: Stored a string variable, varname=%d  vartype=%d, stringlen=%d.\n", varname, vartype, stringlen);
				sttype = 0;
				nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
				if (nstmt == NULL)
					TOKERROR (80, "Invalid subscript")
				stmt = nstmt;

			} else if ((nlp = match_literal (lp, "UPS$")) != NULL) {
				lp = nlp;

				varname = 030;
				vartype = 017;
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nlp = get_source_string (lp, &varname, &vartype, litstring, &stringlen);
				dprintf (DEBUG_TOKEN, "get_source_string: RETURNED:  lp=%x  nlp=%x, varname=%d  vartype=%d,  litstring5='%-.5s' stringlen=%d\n",
						lp, nlp, varname, vartype, litstring, stringlen);
				if (stringlen != 9999) {
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9));
					*stmt++ = ITOWL ((sttype << 9));
					sttype = 1;
					stlen ++;
					*stmt++ = ITOWH (sttype << 9 | (stringlen & 0377));
					*stmt++ = ITOWL (sttype << 9 | (stringlen & 0377));
					memcpy (stmt, litstring, stringlen);
					stlen += (stringlen + 1) / 2;
					stmt += stringlen + (stringlen % 2);
					dprintf (DEBUG_TOKEN, "STRINGIT2: Stored a string literal of length %d\n", stringlen);
					lp = nlp;
				} else if (nlp != lp) {
					lp = nlp;
					stlen ++;
					*stmt++ = ITOWH ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
					*stmt++ = ITOWL ((sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
					dprintf(DEBUG_TOKEN, "STRINGIT2: Stored a string variable, varname=%d  vartype=%d, stringlen=%d.\n", varname, vartype, stringlen);
					sttype = 0;
					nstmt = get_subscript (&lp, &sttype, stmt, &stlen);
					if (nstmt == NULL)
						TOKERROR (80, "Invalid subscript")
					stmt = nstmt;
				}
				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				stlen ++;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);

			}  else if ((nlp = match_literal (lp, "CHR$")) != NULL) {
				lp = nlp;

				varname = 027;
				vartype = 017;
				stlen ++;
				*stmt++ = ITOWH (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));
				*stmt++ = ITOWL (0x8000 | (sttype << 9) | ((varname & 037) << 4) | (vartype & 017));

				sttype = 0;
				varname = 0;
				vartype = 0;

				if (*lp != '(' && *lp != '[')
					TOKERROR (96, "Left parenthesis expected")
				lp = get_next_character (lp + 1);
				sttype = 013;

				nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
				if (stmt == nstmt)
					TOKERROR (34, "Invalid argument to CHR$")
				stmt = nstmt;

				if (*lp != ')' && *lp != ']')
					TOKERROR (96, "Right parenthesis expected")
				lp = get_next_character (lp + 1);

				sttype = 010;
				stlen ++;
				*stmt++ = ITOWH (sttype << 9);
				*stmt++ = ITOWL (sttype << 9);
				sttype = 0;

			} else
				TOKERROR (97, "Invalid source string")

		} else {
			nstmt = ans_formula (&lp, &sttype, stmt, &stlen, &varname, &vartype, &varnum);
			if (stmt == nstmt)
				TOKERROR (34, "Invalid assignment value")
			stmt = nstmt;
		}

		dprintf (DEBUG_TOKEN, "NEW stlen=%d\n", stlen);
		break;
	}

	if (*lp)
		TOKERROR (24, "Characters after statement end")

	dprintf (DEBUG_TOKEN, "STORING STATEMENT LENGTH: %d\n", stlen);
	statement[2] = ITOWH (stlen);			/* Statement len, upper byte */
	statement[3] = ITOWL (stlen);			/* Statement len, lower byte */

	dprintf (DEBUG_TOKEN, "Statement length is: %d words, %dbytes\n", WTOI (statement + 2), WTOI(statement + 2) * 2 + 2);
	Pentry.block = malloc (WTOI(statement + 2) * 2 + 2);
	memcpy(Pentry.block, statement, WTOI(statement + 2) * 2 + 2);
	tape_mach_type = MACH_TYPE_SET_ACCESSB;
	Pentry.type = 0;
	Pentry.startloc = 02000;
	Pentry.filesize = WTOI(statement + 2) + 2;
	Pentry.symtabstart = 02000 + WTOI(statement + 2);
	verbose = 0xff;
	convert = 1;
	convert_program (&Pentry);


#ifdef INTEGRATEDMAIN
                                if (&Pentry.cline != NULL) { 
                                        struct line_table *lptr = &Pentry.cline, *optr;
                                        while (lptr != NULL) {  
                                                if (lptr->data != NULL) {
							if (memcmp(lptr->data, line, strlen(line) > lptr->len ? strlen(line) : lptr->len)) {
								eprintf ("BAD TOKENIZATION, strlen(line)=%d  strlen(cline)=%d:\n",
									strlen(line), lptr->len);
								eprintf ("Original: %s\n", line);
								eprintf ("Tokenize: %.*s\n", lptr->len, lptr->data);
							}
                                                        nfree(lptr->data);
						}
                                                optr = lptr;
                                                lptr = lptr->next;
                                                if (optr != &Pentry.cline) {
                                                        if (optr != NULL)
                                                                nfree(optr);
                                                }
                                        }
                                }

				free(Pentry.block);
#endif /* INTEGRATEDMAIN */

	ps->statement_len = stlen;

/*
int i;
	for (i = 0; i < stlen; i++) {
		eprintf ("Word %i:   %o/%o/%o/%o\n", i, WTOI(&statement[i*2]) >> 15, (WTOI(&statement[i*2]) >> 9) & 0377,
					(WTOI(&statement[i*2]) >> 4 ) & 037, (WTOI(&statement[i*2]) & 017));
	}
*/

	return (0);
		
}

uint8_t *
match_literal (uint8_t *line, uint8_t *literal)
{
	uint32_t	i;
	uint8_t		*nlp;

	nlp = line;
	for (i = 0; i < strlen(literal); i++) {
		nlp = get_next_character(nlp);
		if (!*nlp || *nlp != literal[i])
			return NULL;
		nlp++;
	}
	return (get_next_character(nlp));
}

uint8_t *
match_token (uint8_t * line, uint32_t *opcode, struct BASIC_Tokens *Btok, uint32_t count, uint32_t prio)
{
	int32_t		i, j;
	uint8_t		*nlp;
	uint32_t	toklen;
	uint8_t		*tokstart;
	uint32_t	noskipblanks = prio & 0x8000;

	if (noskipblanks)
		prio = prio & 0x7fff;
#ifdef SEVEREDEBUGGING
	dprintf (DEBUG_TOKEN, "match_token: Looking up item in table, count=%d (%o)\n", count, count);
#endif
	for (i = count - 1; i >= 0; i--) {
		nlp = line;

#ifdef SEVEREDEBUGGING
		dprintf (DEBUG_TOKEN, "match_token: Checking table item #%d (%o)\n", i, i);
#endif
		if (Btok[i].ascii[0] == 0)
			continue;
		if (Btok[i].priority < prio) {
			continue;
		}

#ifdef SEVEREDEBUGGING
		dprintf (DEBUG_TOKEN, "match_token: Comparing %-.3s to %s...\n", line, Btok[i].ascii);
#endif

		tokstart = Btok[i].ascii;
		if (*tokstart == ' ')
			tokstart++;
		toklen = strlen(tokstart);
		if (*(tokstart + toklen - 1) == ' ')
			toklen--;

		for (j = 0; j < toklen; j++) {
			nlp = get_next_character (nlp);
			if (*nlp != *(tokstart + j)) {
				nlp = NULL;
				break;
			}
			nlp++;
		}
		if (nlp != NULL) {
			dprintf (DEBUG_TOKEN, "match_token: MATCH FOUND, tokencode = %o: %s\n", i, Btok[i].ascii);
			*opcode = Btok[i].op;
			return (noskipblanks ? nlp : get_next_character(nlp));
		}
	}
	dprintf (DEBUG_TOKEN, "match_token: NO MATCH FOUND for %-.3s\n", line);
	*opcode = 9999;
	return line;
}

uint8_t *
eval_integer (uint8_t * line, uint32_t *value)
{
	uint8_t		*lp = line;
	uint8_t		ndigits = 0;

	*value = 0;
	while (*lp != 0) {
		lp = get_next_character (lp);
		if (!isdigit(*lp))
			break;
		*value = *value * 10 + (*lp - '0');
		ndigits++;
		lp++;
	}
	if (ndigits)
		return (lp);
	return NULL;
}

uint8_t *
get_next_character (uint8_t *line)
{
	while (*line != 0 && (*line == ' ' || *line == '\t'))
		line++;
	return line;
}

uint8_t *
get_nth_character (uint8_t *line, uint32_t charno)
{
	uint8_t		*lp = line;
	uint32_t	i;
	for (i = 0; i < charno; i++) {
		while (*lp != 0 && (*lp == ' ' || *lp == '\t'))
			i++;
		if (*lp == 0)
			break;
		lp++;
	}
	if (i == charno)
		return (lp);
	return (NULL);
}

#ifdef INTEGRATEDMAIN
int
main(int argc, char **argv)
{
	uint8_t	line[512];
	uint32_t	i = 0;
	if (argc < 2)
		/*debug = 64 + 128 + 256 + 512;*/
		debug = 0;
	else
		debug = 0xffffffff;
	while (1) {
		/*fprintf (stderr, "   ---   Enter a line: ");*/
		if (fgets (line, 511, stdin) == NULL)
			break;
		line[strlen(line) - 1] = 0;
		/*dprintf (DEBUG_INDCONV, "--Line is: %s\n", line);
		eprintf ("\r                             \r%s\n", line);*/

		eprintf (" ---------------------- %d:    %s\n", ++i, line);

#ifndef FOREVER
		if (tokenize (line, NULL))
			exit(1);
#endif

	}
	exit(0);
}
#endif

/*
 * demand_string - If a string constant is found, store the oeprand sttype and the string.
 */

uint32_t
demand_string  (uint8_t *stmt, uint32_t sttype, uint32_t *stlen, uint8_t *lp)
{
	uint32_t	slength;
	uint32_t	quoteval;
	uint32_t	quotevalfound;
	uint8_t		stringline[512];
	uint8_t		*stl;

	lp = get_next_character (lp);
	dprintf (DEBUG_TOKEN, "------ 5 characters at address %x: %-.5s.   5 at %x: %-.5s\n",
				lp, lp + 1);
	stl = stringline;
	slength = 0;
	dprintf (DEBUG_TOKEN, "OLD3 stlen=%d\n", *stlen);
	dprintf (DEBUG_TOKEN, "--- Starting with addr %x, char: %c\n", lp, *lp);
	while (*lp) {
		dprintf (DEBUG_TOKEN, "---    Checking addr %x, char: %c\n", lp, *lp);
		if (*lp == '\'') {		/* single-quoted character 'NN */
			dprintf (DEBUG_TOKEN, "---    Got a single quote at addr %x\n", lp);
			quoteval = 0;
			quotevalfound = 0;
			while (*(lp = get_next_character(lp + 1)) && isdigit(*lp)) {
				quotevalfound++;
				quoteval = quoteval * 10 + *lp - '0';
			}
			if (!quotevalfound) {
				eprintf ("ERROR: Invalid numeric string literal after single-quote.\n");
				return (-1);
			}
			*stl++ = quoteval;
			slength++;
			continue;
		} else if (*lp == '"') {	/* double-quoted string */
			dprintf (DEBUG_TOKEN, "---   Got a double quote at addr %x.\n", lp);
			/*while (*(lp = get_next_character(lp + 1)) && *lp != '"') {*/
			while (*++lp && *lp != '"') {
				dprintf (DEBUG_TOKEN, "---   --- Getting a double-quoted character at addr %x: %c\n",
						lp, *lp);
				*stl++ = *lp;
				slength++;
			}
			if (*lp != '"') {
				eprintf ("ERROR: Unterminated string constant.\n");
				return (-1);
			}
		} else {
			dprintf (DEBUG_TOKEN, "---   NO processing to be done.\n");
			if (slength) {
				break;
			} else {
				dprintf (DEBUG_TOKEN, "NEW3b stlen=%d    slength=%d\n", *stlen, slength);
				return (0);
			}
		}
		if (*lp)
			lp = get_next_character (lp + 1);
	}
	if (slength) {
		*stmt++ = ITOWH (sttype << 9);
		*stmt++ = ITOWL (sttype << 9);
		*stlen += 1;
		*stmt++ = ITOWH ((1 << 9) | slength);
		*stmt++ = ITOWL ((1 << 9) | slength);
		*stlen += 1;
		dprintf (DEBUG_TOKEN, "+++ Copying %d bytes \"%-4.4s\" to statement\n",
				slength, stringline);
		memcpy (stmt, stringline, slength);
		stmt += (slength % 2);
		*stlen += slength / 2 + (slength % 2);
		dprintf (DEBUG_TOKEN, "NEW3c stlen=%d    slength=%d\n", *stlen, slength);
		return (1);
	}
	
	return (0);
}

uint8_t *
get_simple_variable (uint8_t *lp, uint8_t *varname, uint8_t *vartype)
{
	lp = get_next_character (lp);

	*varname = 0;
	*vartype = 0;

	if (!isalpha(*lp)) {
	dprintf (DEBUG_TOKEN, "get_simple_variable:  non-alpha char, not bothering: '%c'\n", *lp);
		return (lp);
	}
	*vartype = 4;
	*varname = toupper(*lp) - '@';
	lp = get_next_character (lp + 1);

	if (isdigit(*lp)) {
		*vartype = 5 + (*lp - '0');
		lp = get_next_character (lp + 1);
	}
	dprintf (DEBUG_TOKEN, "get_simple_variable: varname=%c vartype = %d\n", *varname + '@', *vartype);
	return (lp);
}

uint8_t *
get_string_variable (uint8_t *lp, uint8_t *varname, uint8_t *vartype)
{
	uint8_t		varnum;
	uint8_t		*nlp;
	nlp = get_next_character (lp);

	*varname = 0;
	*vartype = 0;

	if (!isalpha(*nlp))
		return (lp);
	*varname = *nlp;
	nlp = get_next_character (nlp + 1);
	if (isdigit (*nlp)) {
		varnum = *nlp - '0';
		*vartype = 5;
		if (varnum == 0) {
			if (*varname < 'Q') {
				*vartype = *varname - 'A';
				*varname = 033;
			} else {
				*vartype = *varname - 'Q';
				*varname = 034;
			}
		} else if (varnum == 1) {
			if (*varname < 'Q') {
				*vartype = *varname - 'A';
				*varname = 035;
			} else {
				*vartype = *varname - 'Q';
				*varname = 036;
			}
		}
		nlp = get_next_character (nlp + 1);
	} else {
		*varname = *varname - '@';
	}

	if (*nlp != '$') {
		*varname = 0;
		return (lp);
	}
	nlp = get_next_character (nlp + 1);
	dprintf (DEBUG_TOKEN, "get_string_variable: Got variable, varname=%d  vartype=%o\n", *varname, *vartype);
	return (nlp);
}

uint8_t *
get_subscript (uint8_t **lp, uint32_t *sttype, uint8_t *stmt, uint32_t *stlen)
{
	uint8_t		*nlp;
	uint32_t	save_stlen;
	uint8_t		*save_stmt;
	uint8_t		*save_lp;
	uint8_t		*nstmt;
	uint8_t		varname, vartype, varnum;
	uint32_t	parentype = 0;

	parentype = (*stlen & 0x8000) == 0x8000; /* if *stlen biased by 0x8000, uses parenthesis instead of brackets */
	*stlen &= 0x7fff;
	dprintf (DEBUG_TOKEN, " PRE-lp=%x\n", *lp);
	nlp = *lp;
	*lp = get_next_character (*lp);

	if (**lp != '(' && **lp != '[') {
		dprintf (DEBUG_TOKEN, "    PRECHECK-- no left-paren, got '%c'\n", **lp);
		return (stmt);
	}

	*sttype = parentype ? 013 : 012;
	save_stlen = *stlen;
	save_stmt = stmt;
	save_lp = *lp;
	
	*lp = get_next_character (*lp + 1);

	nstmt = ans_formula (lp, sttype, stmt, stlen, &varname, &vartype, &varnum);

	dprintf (DEBUG_TOKEN, "get_subscript: YYY ans_formula returned, **lp=%c  stmt=%x  nstmt=%x\n", **lp, stmt, nstmt);

	if (stmt == nstmt) {
		*sttype = 0;
		*stlen = save_stlen;
		stmt = save_stmt;
		*lp = save_lp;
		*lp = nlp;
		return (stmt);
	}
	stmt = nstmt;

	if (**lp == ',') {
		if (sttype != 0) {
			*stlen += 1;
			*stmt++ = ITOWH ((*sttype) << 9);
			*stmt++ = ITOWL ((*sttype) << 9);
		}
		*lp = get_next_character (*lp + 1);
		*sttype = 002;

		nstmt = ans_formula (lp, sttype, stmt, stlen, &varname, &vartype, &varnum);
		dprintf (DEBUG_TOKEN, "get_subscript: ZZZ ans_formula returned, **lp=%c,  stmt=%x  nstmt=%x\n", **lp);

		if (stmt == nstmt) {
			if (sttype != 0) {
				*stlen += 1;
				*stmt++ = ITOWH ((*sttype) << 9);
				*stmt++ = ITOWL ((*sttype) << 9);
			}
				
			*sttype = 0;
			*stlen = save_stlen;
			stmt = save_stmt;
			*lp = save_lp;
			*lp = nlp;
			return (stmt);
		}
		stmt = nstmt;
	}

	if ((**lp != ')' && **lp != ']')) {
		dprintf (DEBUG_TOKEN, "get_subscript: Missing right parenthesis.\n");
		*sttype = 0;
		*stlen = save_stlen;
		stmt = save_stmt;
		*lp = save_lp;
		return (stmt);
	}

	*sttype = parentype ? 010 : 011;
	*lp = get_next_character (*lp + 1);
		
	dprintf (DEBUG_TOKEN, "get_subscript: GOT A SUBSCRIPT, next position: '%c' -- stlen=%d stmt=%x -- storing sttype=%o\n",
			**lp, *stlen, stmt, *sttype);
	*stlen += 1;
	*stmt++ = ITOWH ((*sttype) << 9);
	*stmt++ = ITOWL ((*sttype) << 9);
	*sttype = 0;
	dprintf (DEBUG_TOKEN, "get_subscript: PUT A SUBSCRIPT, next position: '%c' -- stlen=%d stmt=%x\n", **lp, *stlen, stmt);

	dprintf (DEBUG_TOKEN, "POST-lp=%x\n", *lp);
	return (stmt);
}

uint8_t *
get_source_string (uint8_t *lp, uint8_t *varname, uint8_t *vartype, uint8_t *litstring, uint32_t *stringlen)
{
	uint8_t		*nlp;

	nlp = get_next_character (lp);

	*varname = 0;
	*vartype = 0;
	*stringlen = 9999;

	nlp = get_string_variable (lp, varname, vartype);
	if (lp != nlp) {
		lp = nlp;
		return (nlp);
	}
	/* Look for a string literal */
	if (*lp != '"' && *lp != '\'') {
		varname = 0;
		return (lp);
	}
	nlp = get_string_literal (lp, litstring, stringlen);
	if (lp == nlp) {
		varname = 0;
		return (lp);
	}
	dprintf (DEBUG_TOKEN, "*** Got a string literal, stringlen=%d  litstring=\"%s\"\n", *stringlen, litstring);
	return (nlp);
}

uint8_t *
get_string_literal (uint8_t *lp, uint8_t *litstring, uint32_t *stringlen)
{
	uint8_t		*nlp, *cnlp;
	uint32_t	cval;
	uint32_t	quoteopen = 0;
	uint32_t	stringfound = 0;

	*stringlen = 0;
	nlp = get_next_character (lp);
	while (*nlp) {
		if (*nlp == '\'' && !quoteopen) {
			nlp = get_next_character (nlp + 1);
			cnlp = get_number (nlp, GN_INT, &cval);
			if (cnlp == nlp || cval > 255 || cval < 0) {
				*stringlen = 9999;
				return (lp);
			}
			litstring[*stringlen] = cval;
			*stringlen += 1;
			stringfound++;
			nlp = cnlp;
		} else if (*nlp == '"') {
			if (quoteopen > 0) {
				quoteopen = 0;
				nlp = nlp + 1;
			} else {
				nlp = nlp + 1;
				quoteopen = 1;
				while (*nlp) {
					if (*nlp == '"') {
						nlp = nlp + 1;
						quoteopen = 0;
						stringfound++;
						break;
					}
					litstring[*stringlen] = *nlp;
					*stringlen += 1;
					nlp = nlp + 1;
				}
			}
		} else
			break;
	}
	if (stringfound)
		return (nlp);
	else
		return (lp);
}

uint8_t *
get_relational_operator (uint8_t *lp, uint32_t *sttype)
{
	uint8_t		*nlp, *nlp2;
	nlp = get_next_character (lp);
	if ((nlp2 = match_literal (nlp, "<>")) != NULL) {
		*sttype = 036;
		nlp = nlp2;
	} else if ((nlp2 = match_literal (nlp, ">=")) != NULL) {
		*sttype = 037;
		nlp = nlp2;
	} else if ((nlp2 = match_literal (nlp, "<=")) != NULL) {
		*sttype = 040;
		nlp = nlp2;
	} else if (*nlp == '>') {
		*sttype = 025;
		nlp = get_next_character (nlp + 1);
	} else if (*nlp == '<') {
		*sttype = 026;
		nlp = get_next_character (nlp + 1);
	} else if (*nlp == '#') {
		*sttype = 027;
		nlp = get_next_character (nlp + 1);
	} else if (*nlp == '=') {
		*sttype = 030;
		nlp = get_next_character (nlp + 1);
	} else
		return (lp);
	return (nlp);
}
